"""Top-level package for rebel-forge."""
from __future__ import annotations

from .device import device
from .remote import RemoteConfig, RemoteError, ensure_remote, run_remote_command, sync_project
from . import sample as sample

__all__ = [
    "main",
    "__version__",
    "device",
    "RemoteConfig",
    "RemoteError",
    "ensure_remote",
    "run_remote_command",
    "sync_project",
    "sample",
]

__version__ = "0.10.13"


def main() -> None:
    """Entry point for `python -m rebel_forge` and the console script."""
    from .rebel_forge import main as _forge_main

    _forge_main()



"""Remote execution helpers for delegating compute to Nebius GPUs."""
from __future__ import annotations

import os
import re
import shlex
import shutil
import subprocess
import sys
import time
import uuid
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Mapping, Sequence

_RUN_FLAG = "FORGE_REMOTE_ACTIVE"
_RUN_ID_ENV = "FORGE_REMOTE_RUN_ID"
_PROJECT_ROOT_ENV = "FORGE_REMOTE_PROJECT_ROOT"
_DEFAULT_REMOTE_ROOT = "~/forge_runs"
_DEFAULT_VENV = "~/venvs/rebel-forge"

_DEFAULT_EXCLUDES: tuple[str, ...] = (
    "__pycache__",
    "*.pyc",
    "*.pyo",
    "*.pyd",
    ".git",
    ".hg",
    ".svn",
    ".mypy_cache",
    ".pytest_cache",
    "node_modules",
    "dist",
    "build",
    "*.egg-info",
    ".venv",
    "venv",
    "env",
    ".DS_Store",
)

_ROOT_MARKERS: tuple[str, ...] = (
    "pyproject.toml",
    "setup.cfg",
    "setup.py",
    "requirements.txt",
    "Pipfile",
    "poetry.lock",
    ".git",
    ".hg",
)


class RemoteError(RuntimeError):
    """Raised when Nebius orchestration fails."""


@dataclass
class RemoteConfig:
    """Connection parameters for Nebius remote execution."""

    host: str
    username: str
    port: int = 22
    identity_path: Path | None = None
    known_hosts_path: Path | None = None
    strict_host_key: bool = False
    remote_root: str = _DEFAULT_REMOTE_ROOT
    python: str = "python3"
    venv_path: str | None = _DEFAULT_VENV
    rsync_bin: str = "rsync"
    ssh_bin: str = "ssh"
    extra_excludes: tuple[str, ...] = ()
    passthrough_env: tuple[str, ...] = (
        "PYTHONPATH",
        "HF_HOME",
        "HF_DATASETS_CACHE",
        "HF_TOKEN",
        "HUGGING_FACE_HUB_TOKEN",
        "WANDB_API_KEY",
    )

    @classmethod
    def from_env(
        cls,
        *,
        host: str | None = None,
        username: str | None = None,
        port: int | None = None,
        identity_path: str | None = None,
        remote_root: str | None = None,
        python: str | None = None,
        venv_path: str | None = None,
        strict_host_key: bool | None = None,
    ) -> "RemoteConfig":
        env = os.environ
        host = host or env.get("FORGE_REMOTE_HOST") or env.get("NEBIUS_HOST")
        username = username or env.get("FORGE_REMOTE_USER") or env.get("NEBIUS_USERNAME")
        port_val = (
            port
            or _coerce_int(env.get("FORGE_REMOTE_PORT"))
            or _coerce_int(env.get("NEBIUS_PORT"))
            or 22
        )
        identity_candidate = identity_path or env.get("FORGE_REMOTE_KEY_PATH") or env.get("NEBIUS_KEY_PATH")
        identity_raw = env.get("FORGE_REMOTE_KEY") or env.get("NEBIUS_PRIVATE_KEY")
        venv_path = venv_path or env.get("FORGE_REMOTE_VENV") or env.get("NEBIUS_VENV_PATH") or _DEFAULT_VENV
        remote_root = remote_root or env.get("FORGE_REMOTE_ROOT") or env.get("NEBIUS_REMOTE_ROOT") or _DEFAULT_REMOTE_ROOT
        python = python or env.get("FORGE_REMOTE_PYTHON") or env.get("NEBIUS_PYTHON") or "python3"
        strict_host = strict_host_key if strict_host_key is not None else (
            _coerce_bool(env.get("FORGE_REMOTE_STRICT_HOST"))
            or _coerce_bool(env.get("NEBIUS_STRICT_HOST"))
            or False
        )

        if not host or not username:
            raise RemoteError("Missing Nebius host or username; set FORGE_REMOTE_HOST/USER or NEBIUS_HOST/USERNAME.")

        identity_path_obj = _resolve_identity_file(identity_candidate, identity_raw)
        fingerprint = env.get("FORGE_REMOTE_HOST_FINGERPRINT") or env.get("NEBIUS_HOST_FINGERPRINT")
        known_hosts_path = None
        if fingerprint:
            known_hosts_path = _ensure_known_host(host, port_val, fingerprint)
            strict_host = True

        return cls(
            host=host,
            username=username,
            port=port_val,
            identity_path=identity_path_obj,
            known_hosts_path=known_hosts_path,
            strict_host_key=strict_host,
            remote_root=remote_root,
            python=python,
            venv_path=venv_path,
        )

    def ssh_target(self) -> str:
        return f"{self.username}@{self.host}"

    def base_ssh_args(self) -> list[str]:
        args = [self.ssh_bin, "-p", str(self.port)]
        if self.identity_path:
            args.extend(["-i", str(self.identity_path)])
        if self.known_hosts_path:
            args.extend(["-o", f"UserKnownHostsFile={self.known_hosts_path}"])
        elif not self.strict_host_key:
            args.extend(["-o", "StrictHostKeyChecking=no", "-o", "UserKnownHostsFile=/dev/null"])
        args.extend(["-o", "BatchMode=yes"])
        return args

    def ssh_command(self, remote_command: str) -> list[str]:
        return self.base_ssh_args() + [self.ssh_target(), remote_command]

    def rsync_ssh_transport(self) -> str:
        return " ".join(shlex.quote(arg) for arg in self.base_ssh_args())

    def format_remote_path(self, run_dir: str) -> str:
        root = self.remote_root.rstrip("/")
        run_dir = run_dir.lstrip("/")
        return f"{root}/{run_dir}" if run_dir else root

    def env_passthrough(self) -> dict[str, str]:
        values: dict[str, str] = {}
        for key in self.passthrough_env:
            value = os.environ.get(key)
            if value:
                values[key] = value
        return values


def ensure_remote(
    *,
    project_root: str | os.PathLike[str] | None = None,
    script_path: str | os.PathLike[str] | None = None,
    run_id: str | None = None,
    excludes: Iterable[str] | None = None,
    config: RemoteConfig | None = None,
    sync: bool = True,
    install_requires: bool = False,
) -> bool:
    """Ensure the active script executes on Nebius. Returns True when already remote."""

    if os.environ.get(_RUN_FLAG) == "1":
        return True

    cfg = config or RemoteConfig.from_env()
    _normalize_binaries(cfg)
    script = Path(script_path or sys.argv[0]).resolve()
    if not script.exists():
        raise RemoteError(f"Script not found: {script}")

    root = Path(project_root).resolve() if project_root else _discover_project_root(script)
    try:
        rel_script = script.relative_to(root)
    except ValueError as exc:  # pragma: no cover
        raise RemoteError(f"Script {script} is outside the project root {root}") from exc

    excludes_set = tuple(dict.fromkeys([*_DEFAULT_EXCLUDES, *(excludes or ()), *cfg.extra_excludes]))
    run_slug = run_id or _default_run_id(root)
    remote_project_dir = cfg.format_remote_path(run_slug)

    if sync:
        _sync_project(cfg, root, remote_project_dir, excludes_set)
    if install_requires:
        _install_remote_requirements(cfg, remote_project_dir)

    remote_project_shell = _remote_shell_path(remote_project_dir)
    remote_env: dict[str, str] = {
        _RUN_FLAG: "1",
        _RUN_ID_ENV: run_slug,
        _PROJECT_ROOT_ENV: remote_project_shell,
    }
    remote_env.update(cfg.env_passthrough())

    rel_parent = rel_script.parent
    remote_cwd_shell = remote_project_shell
    if rel_parent != Path("."):
        remote_cwd_shell = f"{remote_project_shell}/{rel_parent.as_posix()}"

    command = [cfg.python, rel_script.name, *sys.argv[1:]]
    exit_code = run_remote_command(cfg, command, remote_cwd=remote_cwd_shell, env=remote_env)
    raise SystemExit(exit_code)


def run_remote_command(
    config: RemoteConfig,
    command: Sequence[str] | str,
    *,
    remote_cwd: str | None = None,
    env: Mapping[str, str] | None = None,
    stream_output: bool = True,
) -> int:
    """Execute a command on Nebius and return its exit status."""

    _normalize_binaries(config)
    remote_segments: list[str] = []
    if config.venv_path:
        venv_shell = _remote_shell_path(config.venv_path)
        remote_segments.append(f'source "{venv_shell}/bin/activate"')
    if remote_cwd:
        shell_cwd = _remote_shell_path(remote_cwd)
        remote_segments.append(f'cd "{shell_cwd}"')

    assignments = []
    if env:
        for key, value in env.items():
            assignments.append(f"{key}={shlex.quote(value)}")
    cmd_str = command if isinstance(command, str) else _join_command(command)
    if assignments:
        cmd_str = f"{' '.join(assignments)} {cmd_str}".strip()
    remote_segments.append(cmd_str)

    remote_command = " && ".join(remote_segments)
    ssh_args = config.ssh_command(remote_command)
    return _run_subprocess(ssh_args, stream_output=stream_output, check=False)


def sync_project(
    *,
    config: RemoteConfig | None = None,
    source: str | os.PathLike[str] | None = None,
    run_id: str | None = None,
    excludes: Iterable[str] | None = None,
) -> str:
    """Synchronize a project tree to Nebius and return the remote path."""

    cfg = config or RemoteConfig.from_env()
    _normalize_binaries(cfg)
    root = Path(source or os.getcwd()).resolve()
    if not root.exists():
        raise RemoteError(f"Source directory does not exist: {root}")
    run_slug = run_id or _default_run_id(root)
    excludes_set = tuple(dict.fromkeys([*_DEFAULT_EXCLUDES, *(excludes or ()), *cfg.extra_excludes]))
    remote_project_dir = cfg.format_remote_path(run_slug)
    _sync_project(cfg, root, remote_project_dir, excludes_set)
    return remote_project_dir


def _sync_project(config: RemoteConfig, root: Path, remote_dir: str, excludes: Iterable[str]) -> None:
    rsync_bin = shutil.which(config.rsync_bin) or shutil.which('/usr/bin/rsync') or shutil.which('/bin/rsync')
    if rsync_bin is None:
        raise RemoteError("rsync is required for remote sync but was not found on PATH")
    remote_shell_dir = _remote_shell_path(remote_dir)
    prep_cmd = config.ssh_command(f'mkdir -p "{remote_shell_dir}"')
    _run_subprocess(prep_cmd)

    rsync_cmd = [rsync_bin, "-az", "--delete"]
    for pattern in excludes:
        rsync_cmd.extend(["--exclude", pattern])
    rsync_cmd.extend([
        "-e",
        config.rsync_ssh_transport(),
        f"{root}/",
        f"{config.ssh_target()}:{remote_shell_dir}/",
    ])
    _run_subprocess(rsync_cmd)


def _install_remote_requirements(config: RemoteConfig, remote_dir: str) -> None:
    requirements_path = _remote_shell_path(f"{remote_dir}/requirements.txt")
    python_prefix = ""
    if config.venv_path:
        venv_shell = _remote_shell_path(config.venv_path)
        python_prefix = f'source "{venv_shell}/bin/activate" && '
    upgrade_pip = f"{python_prefix}pip install --upgrade pip"
    remote = (
        f"({upgrade_pip} || true) && "
        f'if [ -f "{requirements_path}" ]; then '
        f"  {python_prefix}pip install -r \"{requirements_path}\"; "
        f"fi"
    )
    _run_subprocess(config.ssh_command(remote))


def _discover_project_root(path: Path) -> Path:
    current = path if path.is_dir() else path.parent
    for parent in [current, *current.parents]:
        for marker in _ROOT_MARKERS:
            if (parent / marker).exists():
                return parent
    return current


def _default_run_id(root: Path) -> str:
    slug = _slugify(root.name)
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    short = uuid.uuid4().hex[:8]
    return f"{slug}-{timestamp}-{short}"


def _slugify(value: str) -> str:
    value = value.lower()
    value = re.sub(r"[^a-z0-9]+", "-", value)
    value = re.sub(r"-+", "-", value)
    return value.strip("-") or "forge"


def _join_command(command: Sequence[str]) -> str:
    return " ".join(shlex.quote(part) for part in command)


def _coerce_int(value: str | None) -> int | None:
    if value is None:
        return None
    try:
        return int(value)
    except ValueError:
        return None


def _coerce_bool(value: str | None) -> bool:
    if value is None:
        return False
    return value.strip().lower() in {"1", "true", "yes", "on"}



def _normalize_binaries(config: RemoteConfig) -> None:
    config.ssh_bin = _resolve_executable(config.ssh_bin, "ssh")
    config.rsync_bin = _resolve_executable(config.rsync_bin, "rsync")


def _resolve_executable(candidate: str | None, fallback: str) -> str:
    for name in (candidate, fallback, f"/usr/bin/{fallback}", f"/bin/{fallback}"):
        if not name:
            continue
        if name.startswith('/'):
            path = Path(name)
            if path.exists() and path.is_file():
                return str(path)
        else:
            resolved = shutil.which(name)
            if resolved:
                return resolved
    return candidate or fallback


def _resolve_identity_file(candidate: str | None, raw_key: str | None) -> Path | None:
    if candidate:
        candidate_path = Path(candidate).expanduser().resolve()
        if candidate_path.exists():
            _ensure_key_permissions(candidate_path)
            return candidate_path
    repo_key = Path(".nebius_key")
    if repo_key.exists():
        _ensure_key_permissions(repo_key)
        return repo_key.resolve()
    if raw_key and "BEGIN" in raw_key and len(raw_key.strip()) > 80:
        cache_dir = Path.home() / ".cache" / "rebel-forge"
        cache_dir.mkdir(parents=True, exist_ok=True)
        cache_key = cache_dir / "nebius_key"
        if not cache_key.exists() or cache_key.read_text() != raw_key:
            cache_key.write_text(raw_key)
        _ensure_key_permissions(cache_key)
        return cache_key
    default = Path.home() / ".ssh" / "id_rsa"
    if default.exists():
        _ensure_key_permissions(default)
        return default
    return None


def _remote_shell_path(path: str | None) -> str:
    if not path:
        return ""
    if path.startswith("~/"):
        return f"$HOME/{path[2:]}"
    return path


def _ensure_key_permissions(path: Path) -> None:
    try:
        path.chmod(0o600)
    except PermissionError:
        pass


def _ensure_known_host(host: str, port: int, fingerprint: str) -> Path:
    ssh_dir = Path.home() / ".ssh"
    ssh_dir.mkdir(parents=True, exist_ok=True)
    known_hosts = ssh_dir / "rebel-forge_known_hosts"
    fingerprint = fingerprint.strip()
    entry = f"[{host}]:{port} {fingerprint}\n"
    if known_hosts.exists():
        existing = known_hosts.read_text()
        if entry in existing:
            return known_hosts
    with known_hosts.open("a", encoding="utf-8") as handle:
        handle.write(entry)
    return known_hosts


def _run_subprocess(cmd: Sequence[str], *, stream_output: bool = True, check: bool = True) -> int:
    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE if stream_output else subprocess.DEVNULL,
        stderr=subprocess.STDOUT,
        text=True,
    )
    if stream_output and process.stdout is not None:
        for line in process.stdout:
            sys.stdout.write(line)
            sys.stdout.flush()
    code = process.wait()
    if check and code != 0:
        joined = " ".join(shlex.quote(part) for part in cmd)
        raise RemoteError(f"Command failed with exit code {code}: {joined}")
    return code


__all__ = [
    "RemoteConfig",
    "RemoteError",
    "ensure_remote",
    "run_remote_command",
    "sync_project",
]



"""Interactive onboarding banner for the rebel-forge CLI."""
from __future__ import annotations

import json
import os
import textwrap
import time
import uuid
import webbrowser
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional
from urllib import request
from urllib.error import HTTPError, URLError

from .credential_cache import clear_cached_bundle

ANSI_BLUE = "\u001b[38;5;39m"
ANSI_RED = "\u001b[38;5;196m"
ANSI_ACCENT = ANSI_BLUE
ANSI_BOLD = "\u001b[1m"
ANSI_RESET = "\u001b[0m"

ASCII_LOGO = r"""
██████╗ ███████╗██████╗ ███████╗██╗         ███████╗ ██████╗ ██████╗  ██████╗ ███████╗
██╔══██╗██╔════╝██╔══██╗██╔════╝██║         ██╔════╝██╔═══██╗██╔══██╗██╔════╝ ██╔════╝
██████╔╝█████╗  ██████╔╝█████╗  ██║  █████╗ █████╗  ██║   ██║██████╔╝██║  ███╗█████╗  
██╔══██╗██╔══╝  ██╔══██╗██╔══╝  ██║  ╚════╝ ██╔══╝  ██║   ██║██╔══██╗██║   ██║██╔══╝  
██║  ██║███████╗██║  ██║███████╗███████╗    ██║     ╚██████╔╝██║  ██║╚██████╔╝███████╗
╚═╝  ╚═╝╚══════╝╚═╝  ╚═╝╚══════╝╚══════╝    ╚═╝      ╚═════╝ ╚═╝  ╚═╝ ╚═════╝ ╚══════╝
"""

WELCOME_PANEL = r"""
╭──────────────────────────────────────────────────────────────────────────────╮
│ * Welcome to the Rebel Forge research preview!                              │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

INFO_PARAGRAPH = (
    "Rebel Forge is billed based on Nebius usage through your Rebel account.\n"
    "Pricing may evolve as we move towards general availability."
)

CTA_LINE = "Press Enter to sign in to your Rebel account…"


@dataclass(init=False)
class HandshakeSession:
    """Lightweight record describing a CLI unlock session."""

    login_base: str
    token: str
    handshake_path: Path
    portal_url: str
    device_id: str

    def __init__(
        self,
        *,
        login_base: str | None = None,
        token: str,
        handshake_path: Path,
        portal_url: str,
        device_id: str,
        **legacy_kwargs,
    ) -> None:
        base = login_base or legacy_kwargs.pop("base", None)
        if legacy_kwargs:
            unexpected = ", ".join(sorted(legacy_kwargs))
            raise TypeError(f"Unexpected arguments for HandshakeSession: {unexpected}")
        if base is None:
            raise TypeError("login_base is required")
        self.login_base = base
        self.token = token
        self.handshake_path = handshake_path
        self.portal_url = portal_url
        self.device_id = device_id


@dataclass(frozen=True)
class HandshakeRecord:
    """Details returned from the portal handshake."""

    token: str
    user_id: str
    unlocked_at: float
    display_name: str | None = None
    first_name: str | None = None
    last_name: str | None = None
    session_template: str | None = None
    session_id: str | None = None
    device_id: str | None = None


_MARKER_DIR = Path.home() / ".rebel-forge"
_MARKER_PATH = _MARKER_DIR / "onboarding.done"
_SESSION_PATH = _MARKER_DIR / "session.json"
_DEVICE_PATH = _MARKER_DIR / "device.json"
_STATUS_ENDPOINT = "/api/cli-auth/status"
_HANDSHAKE_PREFIX = "cli-handshake-"
_DEFAULT_LOGIN_URL = "http://localhost:3000"
_SKIP_ENV = "REBEL_FORGE_SKIP_ONBOARDING"
_AUTO_UNLOCK_ENV = "REBEL_FORGE_AUTO_UNLOCK"
_AUTO_HANDSHAKE_USER_ENV = "REBEL_FORGE_HANDSHAKE_USER"
_AUTO_HANDSHAKE_NAME_ENV = "REBEL_FORGE_HANDSHAKE_NAME"
_TRUE_VALUES = {"1", "true", "yes", "on"}
_CLEAR_CMDS = ("clear", "cls")
_POLL_ATTEMPTS = 240
_POLL_DELAY_SECONDS = 1.0
_SESSION_VERIFY_TIMEOUT = 5.0
_DEFAULT_HANDSHAKE_USER = "cli-user"


def _load_device_id() -> str | None:
    try:
        raw = _DEVICE_PATH.read_text(encoding="utf-8")
        parsed = json.loads(raw)
    except (FileNotFoundError, json.JSONDecodeError, OSError):
        return None
    value = parsed.get("deviceId") if isinstance(parsed, dict) else None
    if isinstance(value, str):
        candidate = value.strip()
        if candidate:
            return candidate
    return None


def _store_device_id(device_id: str) -> str:
    _MARKER_DIR.mkdir(parents=True, exist_ok=True)
    payload = {
        "deviceId": device_id,
        "createdAt": time.time(),
    }
    _DEVICE_PATH.write_text(json.dumps(payload, indent=2), encoding="utf-8")
    return device_id


def _resolve_device_id() -> str:
    existing = _load_device_id()
    if existing:
        return existing
    return _store_device_id(uuid.uuid4().hex)


def start_handshake_session(*, login_base: str | None = None) -> HandshakeSession:
    base = (login_base or resolve_login_url()).rstrip("/")
    token = uuid.uuid4().hex
    handshake_path = _MARKER_DIR / f"{_HANDSHAKE_PREFIX}{token}"
    device_id = _resolve_device_id()
    portal_url = f"{base}/cli?token={token}&deviceId={device_id}"
    handshake_path.unlink(missing_ok=True)
    return HandshakeSession(
        login_base=base,
        token=token,
        handshake_path=handshake_path,
        portal_url=portal_url,
        device_id=device_id,
    )


def wait_for_handshake(session: HandshakeSession) -> Optional[HandshakeRecord]:
    return _await_handshake(session.portal_url, session.handshake_path, session.token, session.device_id)


def ensure_onboarding(*, workspace: Path) -> Optional[dict[str, Any]]:
    """Guide the user through Rebel account login and return the active session payload when linked."""

    if _should_skip():
        return None

    login_base = resolve_login_url().rstrip("/")
    existing_session = _load_session()

    if existing_session and _session_active(existing_session, login_base):
        _ensure_marker(login_base)
        return existing_session

    if existing_session:
        print("Rebel authentication expired. Please sign in again to unlock the CLI.\n")
        _clear_session_state()

    session = start_handshake_session(login_base=login_base)

    if _auto_unlock_enabled():
        _render_banner(session.portal_url, workspace)
        _launch_login(session.portal_url)
        record = wait_for_handshake(session)
        session.handshake_path.unlink(missing_ok=True)
        if record:
            _record_session(session.login_base, record)
            return _load_session()
        return None

    while True:
        _render_banner(session.portal_url, workspace)
        choice = input("Press Enter to open the Rebel portal, 's' to skip, or 'quit' to exit: ").strip().lower()
        if choice in {"", "login", "sign", "sign in", "signin", "log in", "1"}:
            _launch_login(session.portal_url)
            record = wait_for_handshake(session)
            if record:
                _record_session(session.login_base, record)
                session.handshake_path.unlink(missing_ok=True)
                return _load_session()
            print("Still waiting for confirmation from the browser. Select Link CLI again after signing in.")
            continue
        if choice in {"skip", "s"}:
            if _confirm_skip():
                session.handshake_path.unlink(missing_ok=True)
                return None
            continue
        if choice in {"quit", "q", "exit"}:
            raise SystemExit("Aborted before completing Rebel Forge onboarding.")
        print("Unrecognized option. Press Enter to Log In, type 's' to skip, or 'quit' to exit.")


def update_cli_session(login_base: str, handshake: HandshakeRecord) -> None:
    """Refresh the persisted CLI session after a new portal handshake."""
    _record_session(login_base.rstrip("/"), handshake, quiet=True)


def _should_skip() -> bool:
    return _coerce_bool(os.environ.get(_SKIP_ENV))


def _auto_unlock_enabled() -> bool:
    return _coerce_bool(os.environ.get(_AUTO_UNLOCK_ENV))


def resolve_login_url() -> str:
    for key in ("REBEL_FORGE_LOGIN_URL", "REBEL_FORGE_FRONTEND_URL", "NEXT_PUBLIC_SITE_URL"):
        raw = os.environ.get(key)
        if raw:
            return raw.rstrip("/")
    return _DEFAULT_LOGIN_URL


def _render_banner(portal_url: str, workspace: Path) -> None:
    _clear_screen()
    print(render_welcome_panel())
    print(render_logo())
    print()
    body = textwrap.dedent(
        f"""{INFO_PARAGRAPH}

Workspace ▸ {workspace}
Portal    ▸ {portal_url}
"""
    )
    print(body)
    print(render_cta_line())
    print()


def _clear_screen() -> None:
    if _coerce_bool(os.environ.get("REBEL_FORGE_DISABLE_CLEAR")):
        return
    for cmd in _CLEAR_CMDS:
        if os.system(cmd + " > /dev/null 2>&1") == 0:
            return


def _launch_login(portal_url: str) -> None:
    print(f"Opening {portal_url} in your browser…")
    webbrowser.open(portal_url, new=2, autoraise=True)


def _load_session() -> Optional[dict[str, Any]]:
    try:
        raw = _SESSION_PATH.read_text()
    except FileNotFoundError:
        return None
    except OSError:
        return None
    try:
        payload = json.loads(raw)
    except json.JSONDecodeError:
        return None
    if not isinstance(payload, dict):
        return None
    return payload


def get_cached_session() -> Optional[dict[str, Any]]:
    """Return the cached CLI session payload when present."""

    return _load_session()


def _write_session_payload(payload: dict[str, Any]) -> None:
    _MARKER_DIR.mkdir(parents=True, exist_ok=True)
    _SESSION_PATH.write_text(json.dumps(payload, indent=2))


def _ensure_marker(login_base: str) -> None:
    if _MARKER_PATH.exists():
        return
    _MARKER_DIR.mkdir(parents=True, exist_ok=True)
    payload = {"login_url": login_base, "completed_at": time.time()}
    _MARKER_PATH.write_text(json.dumps(payload, indent=2))


def _clear_session_state() -> bool:
    removed = False
    for path in (_SESSION_PATH, _MARKER_PATH):
        try:
            path.unlink()
        except FileNotFoundError:
            continue
        else:
            removed = True

    try:
        for handshake in _MARKER_DIR.glob(f"{_HANDSHAKE_PREFIX}*"):
            if handshake.is_file():
                handshake.unlink()
                removed = True
    except FileNotFoundError:
        pass

    return removed


def clear_cli_session() -> bool:
    """Public helper to clear onboarding markers and cached credentials."""

    removed = _clear_session_state()
    clear_cached_bundle()
    return removed


def _record_session(login_base: str, handshake: HandshakeRecord, *, quiet: bool = False) -> None:
    clear_cached_bundle()
    device_id = handshake.device_id or _resolve_device_id()
    session_payload = {
        "login_url": login_base,
        "userId": handshake.user_id,
        "token": handshake.token,
        "linked_at": handshake.unlocked_at,
        "last_verified": time.time(),
        "device_id": device_id,
    }
    if handshake.session_template:
        session_payload["session_template"] = handshake.session_template
    if handshake.session_id:
        session_payload["session_id"] = handshake.session_id
    profile: dict[str, str] = {}
    if handshake.display_name:
        session_payload["display_name"] = handshake.display_name
        profile["display_name"] = handshake.display_name
    if handshake.first_name:
        session_payload["first_name"] = handshake.first_name
        profile["first_name"] = handshake.first_name
    if handshake.last_name:
        session_payload["last_name"] = handshake.last_name
        profile["last_name"] = handshake.last_name
    if profile:
        session_payload["profile"] = profile
    _write_session_payload(session_payload)
    _ensure_marker(login_base)
    if not quiet:
        greeting = "Onboarding complete. Welcome to Rebel Forge!"
        if handshake.display_name:
            greeting = f"Onboarding complete. Welcome, {handshake.display_name}!"
        print(f"{greeting}\n")


def _session_active(session_payload: dict[str, Any], login_base: str) -> bool:
    user_id = session_payload.get("userId") or session_payload.get("user_id")
    if not isinstance(user_id, str) or not user_id.strip():
        return False
    base = login_base.rstrip("/")
    url = f"{base}{_STATUS_ENDPOINT}?userId={user_id}"
    request_obj = request.Request(url, headers={"Accept": "application/json"})
    try:
        with request.urlopen(request_obj, timeout=_SESSION_VERIFY_TIMEOUT) as response:
            raw = response.read()
            data = json.loads(raw.decode("utf-8") if raw else "{}")
    except HTTPError as error:
        if error.code in {401, 403, 404}:
            return False
        print(f"[rebel-forge] Warning: CLI status endpoint returned {error.code}; assuming session still active.")
        return True
    except (URLError, TimeoutError, OSError) as error:
        print(f"[rebel-forge] Warning: Unable to reach CLI status endpoint ({error}); assuming session still active.")
        return True
    except Exception as error:  # pragma: no cover - defensive
        print(f"[rebel-forge] Warning: CLI status probe failed ({error}); assuming session still active.")
        return True
    active = bool(data.get("active"))
    if active:
        session_payload["login_url"] = base
        session_payload["last_verified"] = time.time()
        _write_session_payload(session_payload)
    return active


def _await_handshake(portal_url: str, handshake_path: Path, token: str, device_id: str) -> Optional[HandshakeRecord]:
    print("Waiting for the CLI unlock signal from the browser…")
    if _auto_unlock_enabled():
        _write_handshake(handshake_path, token, device_id)
        return _load_handshake_payload(handshake_path, token)
    for attempt in range(1, _POLL_ATTEMPTS + 1):
        if handshake_path.exists():
            return _load_handshake_payload(handshake_path, token)
        if attempt == 1 or attempt % 20 == 0:
            if not is_portal_alive(portal_url):
                print("The portal is not responding yet. Ensure `npm run dev` is active.")
            else:
                print("Portal responding. Complete the Rebel sign-in and wait for the automatic unlock.")
        time.sleep(_POLL_DELAY_SECONDS)
    return None


def _load_handshake_payload(handshake_path: Path, token: str) -> Optional[HandshakeRecord]:
    try:
        payload = json.loads(handshake_path.read_text())
    except (OSError, json.JSONDecodeError):
        return None
    user_raw = payload.get("userId")
    if not isinstance(user_raw, str) or not user_raw.strip():
        return None
    unlocked_raw = payload.get("unlockedAt", time.time())
    try:
        unlocked_at = float(unlocked_raw)
    except (TypeError, ValueError):
        unlocked_at = time.time()
    def _clean(value: object) -> str | None:
        if isinstance(value, str):
            candidate = value.strip()
            if candidate:
                return candidate
        return None

    device_id = _clean(payload.get("deviceId") or payload.get("device_id"))

    raw_profile = payload.get("profile")
    profile = raw_profile if isinstance(raw_profile, dict) else {}
    first_name = _clean(
        profile.get("firstName")
        or profile.get("first_name")
        or payload.get("firstName")
        or payload.get("first_name")
    )
    last_name = _clean(
        profile.get("lastName")
        or profile.get("last_name")
        or payload.get("lastName")
        or payload.get("last_name")
    )
    display_name = _clean(
        profile.get("fullName")
        or profile.get("full_name")
        or profile.get("displayName")
        or payload.get("fullName")
        or payload.get("full_name")
        or payload.get("displayName")
    )
    if not display_name:
        combined = " ".join(part for part in (first_name, last_name) if part)
        display_name = combined.strip() if combined else None

    session_template = _clean(payload.get("sessionTemplate") or payload.get("session_template"))
    session_id = _clean(payload.get("sessionId") or payload.get("session_id"))

    return HandshakeRecord(
        token=token,
        user_id=user_raw.strip(),
        unlocked_at=unlocked_at,
        display_name=display_name,
        first_name=first_name,
        last_name=last_name,
        session_template=session_template,
        session_id=session_id,
        device_id=device_id,
    )


def is_portal_alive(portal_url: str) -> bool:
    base = portal_url.split("?", 1)[0]
    try:
        with request.urlopen(base, timeout=1.5) as response:
            return 200 <= response.status < 500
    except Exception:
        return False


def _write_handshake(handshake_path: Path, token: str, device_id: str) -> None:
    handshake_path.parent.mkdir(parents=True, exist_ok=True)
    profile: dict[str, str] = {}
    auto_name = os.environ.get(_AUTO_HANDSHAKE_NAME_ENV)
    if isinstance(auto_name, str) and auto_name.strip():
        profile["fullName"] = auto_name.strip()
    payload: dict[str, object] = {
        "token": token,
        "userId": os.environ.get(_AUTO_HANDSHAKE_USER_ENV, _DEFAULT_HANDSHAKE_USER),
        "unlockedAt": time.time(),
        "deviceId": device_id,
    }
    if profile:
        payload["profile"] = profile
    handshake_path.write_text(json.dumps(payload, indent=2))


def _confirm_skip() -> bool:
    answer = input("Skip onboarding for this session? [y/N]: ").strip().lower()
    return answer.startswith("y")


def _coerce_bool(value: str | None) -> bool:
    if value is None:
        return False
    return value.strip().lower() in _TRUE_VALUES


def _colorize_block(block: str, color: str, *, bold: bool = False) -> str:
    if os.environ.get("NO_COLOR"):
        return block
    prefix = color + (ANSI_BOLD if bold else "")
    suffix = ANSI_RESET
    return "\n".join(f"{prefix}{line}{suffix}" for line in block.splitlines())


def render_welcome_panel() -> str:
    return _colorize_block(WELCOME_PANEL.strip("\n"), ANSI_ACCENT, bold=True)


def render_cta_line() -> str:
    line = CTA_LINE
    if os.environ.get("NO_COLOR"):
        return line
    return f"{ANSI_BLUE}{ANSI_BOLD}{line}{ANSI_RESET}"


def render_logo(*, force_color: bool = False) -> str:
    art = ASCII_LOGO.strip("\n")
    if force_color or not os.environ.get("NO_COLOR"):
        return _colorize_block(art, ANSI_RED, bold=True)
    return art


__all__ = [
    "ensure_onboarding",
    "clear_cli_session",
    "get_cached_session",
    "resolve_login_url",
    "is_portal_alive",
    "render_logo",
    "render_welcome_panel",
    "render_cta_line",
    "INFO_PARAGRAPH",
    "CTA_LINE",
    "ASCII_LOGO",
    "ANSI_ACCENT",
    "ANSI_BLUE",
    "ANSI_BOLD",
    "ANSI_RED",
    "ANSI_RESET",
    "HandshakeSession",
    "HandshakeRecord",
    "start_handshake_session",
    "wait_for_handshake",
    "update_cli_session",
]



"""Sample remote-training script for rebel-forge users."""
from __future__ import annotations

import socket
from pathlib import Path

import torch
from torch import nn
from torch.utils.data import DataLoader, TensorDataset

try:
    from .remote import ensure_remote
except ImportError:
    from rebel_forge.remote import ensure_remote


def _build_dataset(num_samples: int = 2048) -> TensorDataset:
    generator = torch.Generator().manual_seed(109)
    features = torch.randn(num_samples, 4, generator=generator)
    targets = torch.stack(
        [
            torch.sin(features[:, 0]) + 0.2 * features[:, 1],
            torch.cos(features[:, 2]) - 0.1 * features[:, 3],
        ],
        dim=1,
    )
    return TensorDataset(features, targets)


def _train(device: torch.device) -> None:
    dataset = _build_dataset()
    loader = DataLoader(dataset, batch_size=64, shuffle=True)

    model = nn.Sequential(
        nn.Linear(4, 128),
        nn.ReLU(),
        nn.Linear(128, 2),
    ).to(device)

    optimizer = torch.optim.AdamW(model.parameters(), lr=5e-3)
    loss_fn = nn.MSELoss()

    for epoch in range(5):
        running = 0.0
        for batch_x, batch_y in loader:
            batch_x, batch_y = batch_x.to(device), batch_y.to(device)
            optimizer.zero_grad(set_to_none=True)
            preds = model(batch_x)
            loss = loss_fn(preds, batch_y)
            loss.backward()
            optimizer.step()
            running += loss.item()
        print(f"[remote] epoch {epoch + 1} loss {running / len(loader):.5f}")

    output_dir = Path("runs/forge-demo")
    output_dir.mkdir(parents=True, exist_ok=True)
    torch.save(model.state_dict(), output_dir / "model.pt")
    print(f"[remote] saved checkpoint to {output_dir.resolve()}")


def main() -> None:
    already_remote = ensure_remote()
    if not already_remote:
        return

    if not torch.cuda.is_available():
        raise RuntimeError("Nebius CUDA device required; refusing to run on CPU.")

    device = torch.device("cuda")
    print(f"[remote] running on {socket.gethostname()} with {device}")
    _train(device)


if __name__ == "__main__":
    main()



import importlib
import sys
import tempfile
import types
import unittest
from pathlib import Path
from unittest.mock import patch

sys.path.insert(0, str(Path(__file__).resolve().parents[1] / "src"))

try:
    import torch  # type: ignore
except Exception:
    torch = types.ModuleType("torch")
    torch.device = lambda *_args, **_kwargs: "cpu"  # type: ignore[attr-defined]
    utils_module = types.ModuleType("torch.utils")
    data_module = types.ModuleType("torch.utils.data")
    utils_module.data = data_module  # type: ignore[attr-defined]
    sys.modules["torch"] = torch
    sys.modules["torch.utils"] = utils_module
    sys.modules["torch.utils.data"] = data_module
else:  # pragma: no cover - align torch shims with device tests
    if not hasattr(torch, "device"):
        torch.device = lambda *_args, **_kwargs: "cpu"  # type: ignore[attr-defined]
    if not hasattr(torch, "utils"):
        torch.utils = types.SimpleNamespace()  # type: ignore[attr-defined]
    if not hasattr(torch.utils, "data"):
        torch.utils.data = types.SimpleNamespace()  # type: ignore[attr-defined]


class LogoutCommandTest(unittest.TestCase):
    def test_logout_clears_state_without_portal(self) -> None:
        with tempfile.TemporaryDirectory() as tmp_home:
            temp_home = Path(tmp_home)
            with patch("pathlib.Path.home", return_value=temp_home):
                for module in (
                    "rebel_forge.cli",
                    "rebel_forge.onboarding",
                    "rebel_forge.credential_cache",
                ):
                    sys.modules.pop(module, None)
                cli = importlib.import_module("rebel_forge.cli")
                importlib.import_module("rebel_forge.onboarding")
                cache = importlib.import_module("rebel_forge.credential_cache")
                cache.keyring = None  # type: ignore[attr-defined]

                marker_dir = temp_home / ".rebel-forge"
                marker_dir.mkdir(parents=True, exist_ok=True)
                session_path = marker_dir / "session.json"
                marker_path = marker_dir / "onboarding.done"
                bundle_path = marker_dir / "bundle.json"
                handshake_path = marker_dir / "cli-handshake-testtoken"
                session_path.write_text("{}")
                marker_path.write_text("{}")
                bundle_path.write_text("{}")
                handshake_path.write_text("{}")
                cache._MEMORY_CACHE = {"foo": "bar"}  # type: ignore[attr-defined]

                original_argv = sys.argv
                sys.argv = ["rebel-forge", "logout"]
                try:
                    with patch.object(cli, "_load_cli_environment") as mock_load, patch.object(
                        cli, "_ensure_default_export"
                    ) as mock_export, patch.object(cli, "_ensure_portal_ready") as mock_portal:
                        with self.assertRaises(SystemExit) as exit_info:
                            cli.main()
                finally:
                    sys.argv = original_argv

                mock_load.assert_not_called()
                mock_export.assert_not_called()
                mock_portal.assert_not_called()
                self.assertEqual(exit_info.exception.code, 0)
                self.assertFalse(session_path.exists())
                self.assertFalse(marker_path.exists())
                self.assertFalse(bundle_path.exists())
                self.assertFalse(handshake_path.exists())
                self.assertIsNone(cache.load_cached_bundle())


if __name__ == "__main__":
    unittest.main()



import json
import os
import sys
import tempfile
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

sys.path.insert(0, str(Path(__file__).resolve().parents[1] / "src"))

try:
    import torch  # type: ignore
except Exception:  # pragma: no cover - fallback for test environment
    import types

    torch_module = types.ModuleType("torch")
    utils_module = types.ModuleType("torch.utils")
    data_module = types.ModuleType("torch.utils.data")
    setattr(data_module, "DataLoader", object)
    setattr(data_module, "TensorDataset", object)
    setattr(utils_module, "data", data_module)
    setattr(torch_module, "utils", utils_module)
    setattr(torch_module, "device", lambda *_args, **_kwargs: "cpu")
    sys.modules["torch"] = torch_module
    sys.modules["torch.utils"] = utils_module
    sys.modules["torch.utils.data"] = data_module
else:  # pragma: no cover
    import types
    try:
        import torch.utils.data as torch_utils_data  # type: ignore
    except Exception:
        torch_utils_data = types.SimpleNamespace()
        sys.modules["torch.utils.data"] = torch_utils_data  # type: ignore[assignment]

    if not hasattr(torch, "device"):
        torch.device = lambda *_args, **_kwargs: "cpu"  # type: ignore[attr-defined]
    if not hasattr(torch, "utils"):
        torch.utils = types.SimpleNamespace()  # type: ignore[attr-defined]
    if not hasattr(torch.utils, "data"):
        torch.utils.data = torch_utils_data  # type: ignore[attr-defined]
    if not hasattr(torch_utils_data, "DataLoader"):
        torch_utils_data.DataLoader = object  # type: ignore[attr-defined]
    if not hasattr(torch_utils_data, "TensorDataset"):
        torch_utils_data.TensorDataset = object  # type: ignore[attr-defined]


import importlib

device_mod = importlib.import_module("rebel_forge.device")
from rebel_forge.device import NebiusProvisioner, RemoteError


GPU_PLATFORM_RESPONSE = {
    "spec": {
        "presets": [
            {"name": "1gpu-16vcpu-200gb", "resources": {"gpu_count": 1}},
            {"name": "8gpu-128vcpu-1600gb", "resources": {"gpu_count": 8}},
        ]
    }
}


class AutoPresetTest(unittest.TestCase):
    def test_auto_preset_selects_gpu_count(self) -> None:
        with tempfile.TemporaryDirectory() as tmp_home, patch.dict(
            os.environ, {"HOME": tmp_home}, clear=False
        ):
            provisioner = NebiusProvisioner(
                project_id="project-test",
                service_account_id="sa-test",
                public_key_id="pub-test",
                private_key_pem="-----BEGIN PRIVATE KEY-----\n...\n-----END PRIVATE KEY-----",
                profile="test-profile",
                endpoint="api.nebius.cloud:443",
            )
            with patch.object(
                provisioner,
                "_run_cli",
                return_value=json.dumps(GPU_PLATFORM_RESPONSE),
            ):
                preset = provisioner.auto_preset(platform="gpu-h200-sxm", gpu_count=8)
        self.assertEqual(preset, "8gpu-128vcpu-1600gb")

    def test_auto_preset_rejects_unknown_count(self) -> None:
        with tempfile.TemporaryDirectory() as tmp_home, patch.dict(
            os.environ, {"HOME": tmp_home}, clear=False
        ):
            provisioner = NebiusProvisioner(
                project_id="project-test",
                service_account_id="sa-test",
                public_key_id="pub-test",
                private_key_pem="-----BEGIN PRIVATE KEY-----\n...\n-----END PRIVATE KEY-----",
                profile="test-profile",
                endpoint="api.nebius.cloud:443",
            )
            with patch.object(
                provisioner,
                "_run_cli",
                return_value=json.dumps(GPU_PLATFORM_RESPONSE),
            ):
                with self.assertRaises(RemoteError):
                    provisioner.auto_preset(platform="gpu-h200-sxm", gpu_count=4)

    def test_auto_preset_cpu_defaults_without_cli(self) -> None:
        with tempfile.TemporaryDirectory() as tmp_home, patch.dict(
            os.environ, {"HOME": tmp_home}, clear=False
        ):
            provisioner = NebiusProvisioner(
                project_id="project-test",
                service_account_id="sa-test",
                public_key_id="pub-test",
                private_key_pem="-----BEGIN PRIVATE KEY-----\n...\n-----END PRIVATE KEY-----",
                profile="test-profile",
                endpoint="api.nebius.cloud:443",
            )
            with patch.object(provisioner, "_run_cli", MagicMock()) as mock_cli:
                preset = provisioner.auto_preset(platform="cpu-d3", gpu_count=None)
        self.assertEqual(preset, "8vcpu-32gb")
        mock_cli.assert_not_called()

    def test_resolve_credentials_falls_back_on_remote_error(self) -> None:
        device_mod._CACHED_BUNDLE = None
        env_overrides = {
            "NEBIUS_PROJECT_ID": "project-test",
            "NEBIUS_SERVICE_ACCOUNT_ID": "sa-test",
            "NEBIUS_AUTHORIZED_KEY_ID": "key-test",
            "NEBIUS_AUTHORIZED_PRIVATE_KEY": "-----BEGIN PRIVATE KEY-----\n...\n-----END PRIVATE KEY-----",
            "ssh_key_public": "ssh-rsa dummy rebel@test",
            "ssh_key_private": "-----BEGIN OPENSSH PRIVATE KEY-----\n...\n-----END OPENSSH PRIVATE KEY-----",
        }
        original_load_env = device_mod.load_local_env
        original_session = device_mod._load_session_payload
        original_fetch = device_mod._fetch_remote_bundle
        mock_load_env = MagicMock()
        mock_session = MagicMock(return_value={"token": "tok", "userId": "user"})
        mock_fetch = MagicMock(side_effect=RemoteError("remote error"))
        device_mod.load_local_env = mock_load_env  # type: ignore[assignment]
        device_mod._load_session_payload = mock_session  # type: ignore[assignment]
        device_mod._fetch_remote_bundle = mock_fetch  # type: ignore[assignment]
        env_with_flag = dict(env_overrides)
        env_with_flag["REBEL_FORGE_ALLOW_LEGACY_CREDENTIALS"] = "1"
        try:
            with patch.dict(os.environ, env_with_flag, clear=False):
                bundle = device_mod._resolve_credentials()
            self.assertEqual(bundle["projectId"], "project-test")
            self.assertEqual(bundle["serviceAccountId"], "sa-test")
            mock_load_env.assert_called_once()
            mock_session.assert_called_once()
            mock_fetch.assert_called_once()
        finally:
            device_mod.load_local_env = original_load_env  # type: ignore[assignment]
            device_mod._load_session_payload = original_session  # type: ignore[assignment]
            device_mod._fetch_remote_bundle = original_fetch  # type: ignore[assignment]
            device_mod._CACHED_BUNDLE = None

    def test_normalize_portal_base_adds_scheme(self) -> None:
        self.assertEqual(device_mod._normalize_portal_base("portal.example.com"), "https://portal.example.com")
        self.assertEqual(
            device_mod._normalize_portal_base("https://portal.example.com/"), "https://portal.example.com"
        )
        self.assertIsNone(device_mod._normalize_portal_base(""))

    def test_resolve_portal_base_updates_env(self) -> None:
        session = {"token": "tok", "userId": "user", "login_url": "portal.example.com"}
        env_snapshot = dict(os.environ)
        try:
            with patch.dict(os.environ, {}, clear=True):
                base, updated = device_mod._resolve_portal_base(session)
                self.assertEqual(base, "https://portal.example.com")
                self.assertEqual(updated["login_url"], "https://portal.example.com")
                self.assertEqual(os.environ.get("REBEL_FORGE_PORTAL_URL"), "https://portal.example.com")
        finally:
            os.environ.clear()
            os.environ.update(env_snapshot)

    def test_project_fallback_via_handshake_on_server_error(self) -> None:
        session = {"token": "tok", "userId": "user", "login_url": "https://portal.example.com"}
        env_snapshot = dict(os.environ)
        try:
            with patch.dict(os.environ, {}, clear=True):
                with patch.object(device_mod, "_fetch_default_project_metadata", return_value=None), patch.object(
                    device_mod, "_fetch_handshake_bundle", return_value={
                        "projectId": "project-handshake",
                        "endpoint": "api.nebius.cloud:443",
                        "imageId": "image-abc",
                        "subnetId": "subnet-xyz",
                    }
                ):
                    project_id, updated = device_mod._resolve_project_id(session)
                    self.assertEqual(project_id, "project-handshake")
                    self.assertEqual(updated["projectId"], "project-handshake")
                    self.assertEqual(os.environ.get("NEBIUS_PROJECT_ID"), "project-handshake")
        finally:
            os.environ.clear()
            os.environ.update(env_snapshot)

    def test_server_bootstrap_on_404_then_retry(self) -> None:
        # Prepare a minimal session and environment
        session = {"token": "tok", "userId": "user", "login_url": "https://portal.example.com"}
        with patch.object(device_mod, "_load_session_payload", return_value=session), \
            patch.object(device_mod, "_resolve_project_id", return_value=("project-abc", session)), \
            patch.object(device_mod, "_fetch_handshake_bundle", return_value={
                "serviceAccountId": "sa-1",
                "authorizedKeyId": "key-1",
                "authorizedPrivateKey": "-----BEGIN PRIVATE KEY-----\nabc\n-----END PRIVATE KEY-----",
                "endpoint": "api.nebius.cloud:443",
            }), \
            patch.object(device_mod, "_bootstrap_server_credentials", return_value=True) as mock_bootstrap:

            # Simulate first provision attempt 404 then successful JSON
            def _fake_urlopen(req, timeout=0):
                # Return different responses based on a simple counter
                if not hasattr(_fake_urlopen, "calls"):
                    _fake_urlopen.calls = 0  # type: ignore[attr-defined]
                _fake_urlopen.calls += 1  # type: ignore[attr-defined]
                if _fake_urlopen.calls == 1:  # type: ignore[attr-defined]
                    from urllib.error import HTTPError
                    raise HTTPError(req.full_url, 404, "Not Found", hdrs=None, fp=None)
                class _Resp:
                    status = 200
                    def read(self):
                        return json.dumps({
                            "instance": {
                                "instanceId": "i-1",
                                "host": "203.0.113.11",
                                "username": "ubuntu",
                                "bootDiskId": "d-1",
                                "platform": "gpu-h100-sxm",
                                "preset": "1gpu-16vcpu-200gb",
                                "gpuCount": 1,
                            },
                            "sshPrivateKey": "-----BEGIN OPENSSH PRIVATE KEY-----\n...\n-----END OPENSSH PRIVATE KEY-----",
                            "sshPublicKey": "ssh-rsa abc",
                            "sshFingerprint": "fp:xx",
                        }).encode("utf-8")
                    def __enter__(self):
                        return self
                    def __exit__(self, *args):
                        return False
                return _Resp()

            with patch.object(device_mod.urlrequest, "urlopen", side_effect=_fake_urlopen):
                result = device_mod._server_provision(
                    gpu_type="H100",
                    gpu_count=1,
                    storage_gib=512,
                    preset=None,
                    image_id=None,
                    subnet_id=None,
                    ttl_seconds=None,
                )
                self.assertIsInstance(result, dict)
                self.assertIn("instance", result)
                self.assertIn("ssh_private_key", result)
            mock_bootstrap.assert_called_once()

    def test_device_accepts_string_args_and_fallback(self) -> None:
        with tempfile.TemporaryDirectory() as tmp_home, patch.dict(
            os.environ, {"HOME": tmp_home, "REBEL_FORGE_ALLOW_LEGACY_CREDENTIALS": "1"}, clear=False
        ):
            cli_dir = Path(tmp_home) / ".nebius" / "bin"
            cli_dir.mkdir(parents=True, exist_ok=True)
            (cli_dir / "nebius").write_text("#!/bin/sh\n", encoding="utf-8")

            bundle = {
                "projectId": "project-test",
                "serviceAccountId": "sa-test",
                "authorizedKeyId": "key-test",
                "authorizedPrivateKey": "-----BEGIN PRIVATE KEY-----\nabc\n-----END PRIVATE KEY-----",
                "sshPublicKey": "ssh-rsa dummy rebel@test",
                "sshPrivateKey": "-----BEGIN OPENSSH PRIVATE KEY-----\nxyz\n-----END OPENSSH PRIVATE KEY-----",
            }
            instance = device_mod.NebiusInstance(
                instance_id="instance-123",
                name="forge-instance",
                host="203.0.113.10",
                username="ubuntu",
                disk_id="disk-456",
                platform="gpu-h200-sxm",
                preset="preset-h200",
            )
            provisioner = MagicMock()
            provisioner.auto_preset.side_effect = [
                RemoteError("not found"),
                "preset-h200",
            ]
            provisioner.provision_instance.return_value = instance

            with patch.object(device_mod, "_server_provision", side_effect=RemoteError("offline")), patch.object(
                device_mod, "_resolve_credentials", return_value=bundle
            ), patch.object(
                device_mod, "_wait_for_ssh"
            ), patch.object(device_mod, "ensure_remote", side_effect=SystemExit), patch.object(
                device_mod, "NebiusProvisioner", return_value=provisioner
            ):
                with self.assertRaises(SystemExit):
                    device_mod.device("H100", "2 GPUs", "512 gib")

        self.assertEqual(provisioner.auto_preset.call_count, 2)
        first_call = provisioner.auto_preset.call_args_list[0]
        self.assertEqual(first_call.kwargs["platform"], "gpu-h100-sxm")
        self.assertEqual(first_call.kwargs["gpu_count"], 2)
        second_call = provisioner.auto_preset.call_args_list[1]
        self.assertEqual(second_call.kwargs["platform"], "gpu-h200-sxm")
        self.assertEqual(second_call.kwargs["gpu_count"], 2)
        provisioner.provision_instance.assert_called_once()
        kwargs = provisioner.provision_instance.call_args.kwargs
        self.assertEqual(kwargs["platform"], "gpu-h200-sxm")
        self.assertEqual(kwargs["boot_disk_gib"], 512)

    def test_device_raises_when_remote_unavailable_without_legacy(self) -> None:
        with tempfile.TemporaryDirectory() as tmp_home, patch.dict(
            os.environ, {"HOME": tmp_home}, clear=False
        ):
            with patch.object(
                device_mod, "_server_provision", side_effect=RemoteError("service down")
            ), patch.object(device_mod, "_resolve_credentials") as mock_resolve:
                with self.assertRaises(RemoteError):
                    device_mod.device("H100", "512")
            mock_resolve.assert_not_called()

    def test_resolve_project_id_fetches_metadata(self) -> None:
        session_payload = {
            "token": "token-123",
            "userId": "user-456",
            "login_url": "https://portal.example.com",
        }
        metadata = {
            "projectId": "project-xyz",
            "endpoint": "api.nebius.cloud:443",
            "defaultImageId": "image-abc",
            "defaultSubnetId": "subnet-789",
        }
        original_env = {
            "project_id": os.environ.get("project_id"),
            "NEBIUS_PROJECT_ID": os.environ.get("NEBIUS_PROJECT_ID"),
            "NEBIUS_ENDPOINT": os.environ.get("NEBIUS_ENDPOINT"),
            "NEBIUS_IMAGE_ID": os.environ.get("NEBIUS_IMAGE_ID"),
            "NEBIUS_SUBNET_ID": os.environ.get("NEBIUS_SUBNET_ID"),
        }
        try:
            with patch.object(device_mod, "_fetch_default_project_metadata", return_value=metadata):
                project_id, updated = device_mod._resolve_project_id(session_payload)
            self.assertEqual(project_id, "project-xyz")
            self.assertEqual(updated["projectId"], "project-xyz")
            self.assertEqual(os.environ.get("project_id"), "project-xyz")
            self.assertEqual(os.environ.get("NEBIUS_PROJECT_ID"), "project-xyz")
            self.assertEqual(os.environ.get("NEBIUS_ENDPOINT"), "api.nebius.cloud:443")
            self.assertEqual(os.environ.get("NEBIUS_IMAGE_ID"), "image-abc")
            self.assertEqual(os.environ.get("NEBIUS_SUBNET_ID"), "subnet-789")
        finally:
            for key, value in original_env.items():
                if value is None:
                    os.environ.pop(key, None)
                else:
                    os.environ[key] = value


if __name__ == "__main__":
    unittest.main()
